""" Command Line Interface Module """
import optparse
import sys
import os
import yaml
import re
import json
import hashlib
from pymongo import MongoClient
from bandicoot.restapi import routes
from bandicoot.plugins import builtins
from bandicoot.exceptions import DecryptWrongKeyException, DecryptNotClearTextException, DecryptException
from Crypto.Cipher import AES
from hashlib import md5
import binascii
from jinja2 import Template
import ssl
import logging
from logging.handlers import RotatingFileHandler
import time
from glob import glob
import shutil
import datetime
import multiprocessing
import copy


db = None
encryption_password = None
ldap_server = None
ldap_use_ssl = True
ldap_user_cn = None


def counters_db_init(name):
    result = db.counters.find_one( {"_id": name} )
    if result is None:
        db.counters.insert_one( {"_id": name, "seq": 0 } )


def counters_db_getNextSequence(name):
   ret = db.counters.update_one({ "_id": name },
                                { "$inc": { "seq": 1 } },
                                )
   result = db.counters.find_one( {"_id": name} )
   return result["seq"]


def schedule_manager():
    schedule_last_run = {} # stores last run times for 
    global db

    # Setup DB Connection For Thread
    db = MongoClient('localhost').bandicoot

    while True:
        # Get Current Time
        cron_minute = datetime.datetime.now().minute
        cron_hour = datetime.datetime.now().hour
        cron_day_of_month = datetime.datetime.today().day
        cron_month = datetime.datetime.today().month
        cron_day_of_week = datetime.datetime.today().weekday()

        cursor = db.schedules.find()
        for doc in list(cursor):
            name = doc["name"]
            user = doc["user"]
            category = doc["category"]
            action = doc["action"]
            options = None # Default (NOT WORKING)
            minute = "*" # Default
            hour = "*" # Default
            day_of_month = "*" # Default
            month = "*" # Default
            day_of_week = "*" # Default

            if "options" in doc:
                options = doc["options"]
            if "minute" in doc:
                minute = doc["minute"]
            if "hour" in doc:
                hour = doc["hour"]
            if "day_of_month" in doc:
                day_of_month = doc["day_of_month"]
            if "month" in doc:
                month = doc["month"]
            if "day_of_week" in doc:
                day_of_week = doc["day_of_week"]

            # * matches anything, so make it match the current time
            if minute == "*":
                minute = cron_minute
            else:
                minute = int(minute)

            if hour == "*":
                hour = cron_hour
            else:
                hour = int(hour)

            if day_of_month == "*":
                day_of_month = cron_day_of_month
            else:
                day_of_month = int(day_of_month)

            if month == "*":
                month = cron_month
            else:
                month = int(month)

            if day_of_week == "*":
                day_of_week = cron_day_of_week
            else:
                day_of_week = int(day_of_week)

            # Check if cron should be run, see if each setting matches
            # If name is not in schedule_last_run its the first time running it, so thats ok.
            # If name is already in schedule_last_run then check to make sure it didnt already run within the same minute
            if cron_minute == minute and \
               cron_hour == hour and \
               cron_day_of_month == day_of_month and \
               cron_month == month and \
               cron_day_of_week == day_of_week and \
               (name not in schedule_last_run or \
               not (schedule_last_run[name][0] == cron_minute and \
               schedule_last_run[name][1] == cron_hour and \
               schedule_last_run[name][2] == cron_day_of_month and \
               schedule_last_run[name][3] == cron_month and \
               schedule_last_run[name][4] == cron_day_of_week)):

                # Run Scheduled Action
                dat = parse_action(user, category, action, options)

                # Audit Logging / History
                log_action(user, {"result": dat, "category": category, "action": action, "options": options})

                schedule_last_run[name] = (cron_minute, cron_hour, cron_day_of_month, cron_month, cron_day_of_week)

        # Delay 10 seconds between each check
        time.sleep(10)


plugins = {}

builtin_actions = [{'category': '/actions', 'plugin': 'actions_list', 'action': 'list', 'desc': 'list actions'},
                  {'category': '/actions', 'plugin': 'actions_del', 'action': 'del', 'desc': 'del actions'},
                  {'category': '/actions', 'plugin': 'actions_edit', 'action': 'edit', 'desc': 'edit actions'},
                  {'category': '/actions', 'plugin': 'actions_add', 'action': 'add', 'desc': 'add actions'},
                  {'category': '/users', 'plugin': 'users_list', 'action': 'list', 'desc': 'list users'},
                  {'category': '/users', 'plugin': 'users_del', 'action': 'del', 'desc': 'del users'},
                  {'category': '/users', 'plugin': 'users_edit', 'action': 'edit', 'desc': 'edit users'},
                  {'category': '/users', 'plugin': 'users_add', 'action': 'add', 'desc': 'add users'},
                  {'category': '/roles', 'plugin': 'roles_list', 'action': 'list', 'desc': 'list roles'},
                  {'category': '/roles', 'plugin': 'roles_del', 'action': 'del', 'desc': 'del roles'},
                  {'category': '/roles', 'plugin': 'roles_edit', 'action': 'edit', 'desc': 'edit roles'},
                  {'category': '/roles', 'plugin': 'roles_add', 'action': 'add', 'desc': 'add roles'},
                  {'category': '/secrets', 'plugin': 'secrets_list', 'action': 'list', 'desc': 'list secrets'},
                  {'category': '/secrets', 'plugin': 'secrets_del', 'action': 'del', 'desc': 'del secrets'},
                  {'category': '/secrets', 'plugin': 'secrets_edit', 'action': 'edit', 'desc': 'edit secrets'},
                  {'category': '/secrets', 'plugin': 'secrets_add', 'action': 'add', 'desc': 'add secrets'},
                  {'category': '/secrets', 'plugin': 'secrets_encryptpw', 'action': 'encryptpw', 'desc': 'Change password encryption'},
                  {'category': '/plugins', 'plugin': 'plugins_list', 'action': 'list', 'desc': 'list plugins'},
                  {'category': '/', 'plugin': 'ping', 'action': 'ping', 'desc': 'verify connectivity'},
                  {'category': '/', 'plugin': 'logs', 'action': 'logs', 'desc': 'show the history log'},
                  {'category': '/', 'plugin': 'help', 'action': 'help', 'desc': 'print usage'},
                  {'category': '/help', 'plugin': 'help', 'action': '*', 'desc': 'print usage'},
                  {'category': '/jobs', 'plugin': 'jobs_list', 'action': 'list', 'desc': 'list jobs'},
                  {'category': '/jobs', 'plugin': 'jobs_status', 'action': 'status', 'desc': 'get status of job'},
                  {'category': '/jobs', 'plugin': 'jobs_kill', 'action': 'kill', 'desc': 'kill a job'},
                  {'category': '/schedules', 'plugin': 'schedules_add', 'action': 'add', 'desc': 'add schedule'},
                  {'category': '/schedules', 'plugin': 'schedules_edit', 'action': 'edit', 'desc': 'edit schedule'},
                  {'category': '/schedules', 'plugin': 'schedules_list', 'action': 'list', 'desc': 'list schedules'},
                  {'category': '/schedules', 'plugin': 'schedules_del', 'action': 'del', 'desc': 'del schedule'},
                  {'category': '/inventory', 'plugin': 'inventory_list', 'action': 'list', 'desc': 'list inventory'},
                  {'category': '/inventory', 'plugin': 'inventory_del', 'action': 'del', 'desc': 'del inventory item'},
                  {'category': '/', 'plugin': 'stats', 'action': 'stats', 'desc': 'statistics'},
                  ]


def load_plugins(plugin_paths=None):
    default_plugin_path = os.path.dirname(os.path.realpath(__file__)) + "/../plugins/"
    if plugin_paths is None:
        plugin_paths = default_plugin_path
    for plugin_path in plugin_paths.split(":"):
        sys.path.append(plugin_path)
        for file in os.listdir(plugin_path):
            if file.endswith(".py") and file != "__init__.py":
                plugin_module = __import__(file.rstrip(".py"), fromlist=[''])
                load_plugins_from_module(plugin_module)


def load_plugins_from_module(module):
    import inspect
    global plugins
    for member in inspect.getmembers(module):
        plugin_name = member[0]
        plugin_function = member[1]
        if inspect.isfunction(plugin_function):
            if plugin_name not in plugins:
                m = re.match(r'^plugin_(.*?)$', plugin_name)
                if m:
                    plugin_short_name = m.group(1)
                    plugins[plugin_short_name] = plugin_function


def log_action(username, post):
    if post["category"] is not None and post["action"] is not None:
        if post["options"] is not None:
            # Filter sensitive information from options
            for option in ["password", "secret"]:
                if option in post["options"]:
                    post["options"][option] = "..."
        # Only Log Valid Requests
        post["date"] = datetime.datetime.utcnow()
        post["user"] = username
        db.logs.insert_one(post)


def encrypt_dict(dictobj):
    # encrypt sensitive option vals
    global encryption_password
    for key in ["secret"]:
        if dictobj is not None and key in dictobj:
            dictobj[key] = encrypt_str(dictobj[key], encryption_password)
    return True


def decrypt_dict(dictobj):
    # decrypt sensitive option vals
    global encryption_password
    for key in ["secret"]:
        if dictobj is not None and key in dictobj:
            try:
                decrypted_str = decrypt_str(dictobj[key], encryption_password, keyname=dictobj["name"])
                dictobj[key] = decrypted_str
            except DecryptException:
                return False
    return True


def aes_derive_key_and_iv(password, salt, key_length, iv_length):
    """ source: Ansible source code """
    """ Create a key and an initialization vector """
    d = d_i = ''
    while len(d) < key_length + iv_length:
        text = ''.join([d_i, password, salt])
        d_i = str(md5(text).digest())
        d += d_i
    key = d[:key_length]
    iv = d[key_length:key_length+iv_length]
    return key, iv


def encrypt_str(text, encrypt_password=None, key_len=32, encryption_prefix="__bandicoot_encrypted__:"):
    global encryption_password
    if encrypt_password is None and encryption_password is not None:
        # If No encryption password provided, use global encryption password
        encrypt_password = encryption_password
    encrypt_text = encryption_prefix + text
    if encrypt_password is not None:
        salt = "__Salt__"
        key, iv = aes_derive_key_and_iv(encrypt_password, salt, key_len, AES.block_size)
        encryption_suite = AES.new(key, AES.MODE_CFB, iv)
        return str(binascii.b2a_base64(encryption_suite.encrypt(encrypt_text)))
    return str(encrypt_text)


def decrypt_str(text, encrypt_password=None, key_len=32, encryption_prefix="__bandicoot_encrypted__:", keyname="unknown"):
    global encryption_password
    if encrypt_password is None and encryption_password is not None:
        # If No encryption password provided, use global encryption password
        encrypt_password = encryption_password
    if text[:len(encryption_prefix)] == encryption_prefix:
        # Clear Text, No Encryption Password Provided
        return str(text[len(encryption_prefix):])
    elif encrypt_password is not None:
        # Decrypt using password
        salt = "__Salt__"
        key, iv = aes_derive_key_and_iv(encrypt_password, salt, key_len, AES.block_size)
        decryption_suite = AES.new(key, AES.MODE_CFB, iv)
        decrypt_text = str(decryption_suite.decrypt(binascii.a2b_base64(text)))
        if decrypt_text[:len(encryption_prefix)] == encryption_prefix:
            # Decrypted Text
            return str(decrypt_text[len(encryption_prefix):]) 
        else:
            # Probably Wrong Key
            raise DecryptWrongKeyException("  error: Failed to decrypt a secret named %s. If you recently changed your encryption_password try 'secrets encryptpw oldpw=XXXX newpw=XXXX'." % keyname)
    else:
        # Decryption Failed, Its Not Clear Text
        raise DecryptNotClearTextException("  error: Failed to decrypt a secret named %s. If you recently disabled your encryption_password then re-enable it." % keyname)


def secret_has_permission(user, secret):
    cursor = db.roles.find()
    for doc in list(cursor):
        if user in list(doc["users"].split(",")):
            if "secrets" not in doc:
                # No secrets option, give them access to all secrets
                return True
            if secret in list(doc["secrets"].split(",")):
                return True
    return False


def roles_has_permission(user, action, options):
    # Ping is always allowed
    if action["category"] == "/" and action["action"] == "ping":
        return True
    # Help is always allowed
    if action["category"] == "/" and action["action"] == "help":
        return True
    # jobs status is always allowed
    if action["category"] == "/jobs" and action["action"] == "status":
        return True
    # jobs list is always allowed
    if action["category"] == "/jobs" and action["action"] == "list":
        return True
    # jobs kill is always allowed
    if action["category"] == "/jobs" and action["action"] == "kill":
        return True
    """ This allows users to edit their own password.
     sers edit password is allowed if the username is only changing their own password.
     If username is not in options, that means their changing their own password. """
    if action["category"] == "/users" and action["action"] == "edit" and ("username" not in options or options["username"] == user):
        return True

    if action["category"][-1:] == "/":
        action_str = "%s%s" % (action["category"], action["action"])
    else:
        action_str = "%s/%s" % (action["category"], action["action"])
    cursor = db.roles.find()
    for doc in list(cursor):
        if user in list(doc["users"].split(",")):
            for action in list(doc["actions"].split(",")):
                if re.match(r"^%s" % action, action_str):
                    return True
    return False


def clean_all_secrets():
    if not os.path.isdir("/tmp/bandicoot/"):
        os.mkdir("/tmp/bandicoot")

    # Make sure directory permissions are secure
    os.chmod("/tmp/bandicoot/", 0700)

    for filename in glob("/tmp/bandicoot/*"):
        if os.path.isdir(filename):
            shutil.rmtree(filename)
        else:
            os.remove(filename)


def clean_secrets(secrets):
    if secrets is None:
        return None

    for filename in secrets:
        # Temp File must exist
        if os.path.isfile(filename):
            # Delete secret files
            os.remove(filename)


def render_secret_file(name, secret):
    filepath = "/tmp/bandicoot/"
    filename = "%s.%s" % (name, time.time())
    fullpath = "%s%s" % (filepath, filename)

    if not os.path.isdir(filepath):
        os.mkdir(filepath)

    with open(fullpath, "w") as textfile:
        textfile.write(secret)

    os.chmod(fullpath, 0700)

    return fullpath


def render_secrets(user, dictobj):
    secrets = {}
    tmp_secret_files = []

    if dictobj is None or user is None:
        return None

    cursor = db.secrets.find()
    for doc in list(cursor):
        res = decrypt_dict(doc)
        if res is False:
            # Decryption Failed
            return None
        if secret_has_permission(user, doc["name"]):
            if "type" in doc and doc["type"] == "file":
                secrets[doc["name"]] = render_secret_file(doc["name"], doc["secret"])
                tmp_secret_files.append(secrets[doc["name"]])
            else:
                secrets[doc["name"]] = doc["secret"]

    render_vars("secret", secrets, dictobj)
    return tmp_secret_files


def render_vars(varname, vardict, dictobj):
    if dictobj is None or vardict is None:
        return None

    for key in dictobj:
        if isinstance(dictobj[key], basestring):
            t = Template(dictobj[key])
            dictobj[key] = t.render({varname: vardict})


def parse_action(user, category, action, options):
    cursor = db.actions.find()
    for dbaction in builtin_actions + list(cursor):
        if dbaction["category"] == category and (dbaction["action"] == action or dbaction["action"] == "*"):
            new_dbaction = copy.copy(dbaction) # Make a copy to prevent modifying global builtin_actions
            new_dbaction["action"] = action
            if "plugin" in dbaction:
                if not roles_has_permission(user, dbaction, options):
                    return json.dumps({"response": "  you do not have permission to run this action"})
                else:
                    # Admin functions do not allow secrets
                    if dbaction["category"] not in ["/actions", "/users", "/roles", "/secrets", "/plugins"]:
                        if dbaction["category"] == "/" and dbaction["action"] in ["ping"]:
                            # /ping does not allow secrets
                            pass
                        else:
                            # Run Plugin With Secret
                            render_vars("option", options, dbaction)
                            tmp_files_dbaction = render_secrets(user, dbaction)
                            tmp_files_options = render_secrets(user, options)

                            # Check Decryption Failed
                            if user is not None:
                                if (dbaction is not None and tmp_files_dbaction is None) or (options is not None and tmp_files_options is None):
                                    return json.dumps({"response": "  error: Failed to decrypt a secret. If you recently changed your encryption_password try 'secrets encryptpw oldpw=XXXX newpw=XXXX'."})

                            response = plugins[dbaction["plugin"]](user, new_dbaction, options)
                            response = json.loads(response)
                            clean_secrets(tmp_files_dbaction)
                            clean_secrets(tmp_files_options)

                            # async, return queue_id
                            if "response" not in response:
                                if "queue_id" not in response:
                                    return json.dumps({"response": "  error: expected async queue id but found none"})

                            return json.dumps(response)
                    # Run Plugin Without Secrets
                    return plugins[dbaction["plugin"]](user, new_dbaction, options)
    return None


class Cli(object):
    """ bandicoot CLI """

    def __init__(self):
        """ Setup Arguments and Options for CLI """
        # Parse CLI Arguments
        parser = optparse.OptionParser()
        parser.add_option("-s", "--server", dest="server",
                          help="IP address or hostname of bandicoot-api server",
                          metavar="SERVER",
                          default=None)
        parser.add_option("-p", "--port", dest="port",
                          help="tcp port of bandicoot-api server",
                          metavar="PORT",
                          default=None)
        parser.add_option("-t", "--insecure", dest="is_secure",
                          help="Do Not Use SSL",
                          metavar="SECURE",
                          action="store_false",
                          default=True)
        parser.add_option("-d", "--debug", dest="is_debug",
                          help="Debug Mode",
                          metavar="DEBUG",
                          action="store_true",
                          default=False)
        parser.add_option("-k", "--ssl_key", dest="ssl_key",
                          help="SSL key",
                          metavar="SSLKEY",
                          default=None)
        parser.add_option("-c", "--ssl_crt", dest="ssl_crt",
                          help="SSL certificate",
                          metavar="SSLCRT",
                          default=None)
        parser.add_option("-l", "--ldap_server", dest="ldap_server",
                          help="LDAP Server for Authentiation",
                          metavar="LDAPSERVER",
                          default=None)
        parser.add_option("-z", "--ldap_use_ssl", dest="ldap_use_ssl",
                          help="Enable SSL for LDAP",
                          metavar="LDAPUSESSL",
                          default=None)
        parser.add_option("-x", "--ldap_user_cn", dest="ldap_user_cn",
                          help="LDAP User CN",
                          metavar="LDAPUSERCN",
                          default=None)
        parser.add_option("-P", "--pluginpath", dest="pluginpath",
                          help="Plugin Paths, seperated by :",
                          metavar="PLUGINPATH",
                          default=None)
        global encryption_password
        global ldap_server
        global ldap_use_ssl
        global ldap_user_cn
        (options, args) = parser.parse_args()
        self.server = options.server
        self.port = options.port
        self.is_secure = options.is_secure
        self.is_debug = options.is_debug
        self.ssl_key = options.ssl_key
        self.ssl_crt = options.ssl_crt
        ldap_server = options.ldap_server
        ldap_use_ssl = options.ldap_use_ssl
        ldap_user_cn = options.ldap_user_cn
        pluginpath = options.pluginpath

        # Assign values from conf
        bandicoot_config_locations = [os.path.expanduser("~")+"/.bandicoot-api.conf", "/etc/bandicoot-api.conf"]
        bandicoot_conf_obj = {}
        for bandicoot_conf in bandicoot_config_locations:
            if os.path.isfile(bandicoot_conf):
                with open(bandicoot_conf, 'r') as stream:
                    try:
                        bandicoot_conf_obj = yaml.load(stream)
                    except yaml.YAMLError as excep:
                        print("%s\n" % excep)
        if self.server is None and "server" in bandicoot_conf_obj:
            self.server = str(bandicoot_conf_obj["server"])
        if self.port is None and "port" in bandicoot_conf_obj:
            self.port = int(bandicoot_conf_obj["port"])
        if self.is_secure == True and "secure" in bandicoot_conf_obj:
            self.is_secure = bool(bandicoot_conf_obj["secure"])
        if self.is_debug == True and "debug" in bandicoot_conf_obj:
            self.is_debug = bool(bandicoot_conf_obj["debug"])
        if encryption_password is None and "encryption_password" in bandicoot_conf_obj:
            encryption_password = str(bandicoot_conf_obj["encryption_password"])
        if self.ssl_key == None and "ssl_key" in bandicoot_conf_obj:
            self.ssl_key = bool(bandicoot_conf_obj["ssl_key"])
        if self.ssl_crt == None and "ssl_crt" in bandicoot_conf_obj:
            self.ssl_crt = bool(bandicoot_conf_obj["ssl_crt"])
        if ldap_server == None and "ldap_server" in bandicoot_conf_obj:
            ldap_server = str(bandicoot_conf_obj["ldap_server"])
        if ldap_use_ssl == None and "ldap_use_ssl" in bandicoot_conf_obj:
            ldap_use_ssl = str(bandicoot_conf_obj["ldap_use_ssl"])
        if ldap_user_cn == None and "ldap_user_cn" in bandicoot_conf_obj:
            ldap_user_cn = str(bandicoot_conf_obj["ldap_user_cn"])
        if pluginpath == None and "pluginpath" in bandicoot_conf_obj:
            pluginpath = str(bandicoot_conf_obj["pluginpath"])

        # Assign Default values if they were not specified at the cli or in the conf
        if self.server is None:
            self.server = "127.0.0.1"
        if self.port is None:
            self.port = 8088
        if self.ssl_key is None:
            self.ssl_key = "/usr/local/etc/openssl/certs/bandicoot.key"
        if self.ssl_crt is None:
            self.ssl_crt = "/usr/local/etc/openssl/certs/bandicoot.crt"
        if ldap_server is None:
            ldap_server = options.ldap_server
        if ldap_use_ssl is None:
            ldap_use_ssl = options.ldap_use_ssl
        if ldap_user_cn is None:
            ldap_user_cn = options.ldap_user_cn

        # Load Plugins
        load_plugins(pluginpath)

        # Clean any left over secret files
        clean_all_secrets()

    def run(self):
        """ EntryPoint Of Application """
        global db

        # Setup logging to logfile (only if the file was touched)
        if os.path.isfile("/var/log/bandicoot.log"):
            handler = RotatingFileHandler('/var/log/bandicoot.log', maxBytes=10000, backupCount=1)

        # Assign Default values if they were not specified at the cli or in the conf
        if self.server is None:
            self.server = "127.0.0.1"
        if self.port is None:
            self.port = 8088
        if self.ssl_key is None:
            self.ssl_key = "/usr/local/etc/openssl/certs/bandicoot.key"
        if self.ssl_crt is None:
            self.ssl_crt = "/usr/local/etc/openssl/certs/bandicoot.crt"
        if ldap_server is None:
            ldap_server = options.ldap_server
        if ldap_use_ssl is None:
            ldap_use_ssl = options.ldap_use_ssl
        if ldap_user_cn is None:
            ldap_user_cn = options.ldap_user_cn

        # Load Plugins
        load_plugins(pluginpath)

        # Clean any left over secret files
        clean_all_secrets()

    def run(self):
        """ EntryPoint Of Application """
        global db

        # Setup logging to logfile (only if the file was touched)
        if os.path.isfile("/var/log/bandicoot.log"):
            handler = RotatingFileHandler('/var/log/bandicoot.log', maxBytes=10000, backupCount=1)
            handler.setLevel(logging.INFO)
            routes.app.logger.addHandler(handler)
            # Disable stdout logging since its logging to a log file
            log = logging.getLogger('werkzeug')
            log.disabled = True

        # First Time Defaults, Setup superadmin if it doesnt exist
        default_user = "superadmin"
        default_password = "superadmin"
        default_role = "super"

        # Start Scheduler
        p = multiprocessing.Process(target=schedule_manager)
        p.start()

        # Setup DB Connection
        db = MongoClient('localhost').bandicoot

        # Init db counters for jobs
        counters_db_init("jobid")

        # Create default user
        post = db.users.find_one({"username": default_user})
        if post is None:
            m = hashlib.md5()
            m.update(default_password)
            password_md5 = str(m.hexdigest())
            post = {"username": default_user, "password_md5": password_md5}
            db.users.insert_one(post)

        # Create default role
        post = db.roles.find_one({"name": default_role})
        if post is None:
            post = {"name": default_role, "users": default_user, "actions": "/"}
            db.roles.insert_one(post)

        # Start API Server
        routes.app.logger.info("Starting bandicoot api server on %s://%s:%d" % ("https" if
            self.is_secure else "http", self.server, self.port))
        if self.is_secure:
            context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
            context.check_hostname = False
            context.load_cert_chain(certfile=self.ssl_crt, keyfile=self.ssl_key)
            routes.app.run(threaded=True, host=self.server, ssl_context=context, port=self.port, debug=self.is_debug)
        else:
            routes.app.run(threaded=True, host=self.server, port=self.port, debug=self.is_debug)






from bandicoot.parser.lex import *
import ply.yacc as yacc


parser_category = None
parser_action = None
parser_options = None
parser_error = None

# YACC parser
precedence = (
    ('left', 'SPACE'),
    )

def p_action_run(t):
    '''action_run : actions SPACE options
              | actions'''
    global parser_category
    global parser_action
    global parser_options
    if len(t) == 4:
        parser_category = "/%s" % "/".join(t[1][:-1])
        parser_action = t[1][-1]
        parser_options = t[3]
    elif len(t) == 2:
        parser_category = "/%s" % "/".join(t[1][:-1])
        parser_action = t[1][-1]

def p_actions(t):
    '''actions : actions SPACE ACTION
              | ACTION'''
    if t[0] is None:
        t[0] = []
    if len(t) == 4:
        if isinstance(t[1], list):
            t[0] += t[1]
        else:
            t[0].append(t[1])
        t[0].append(t[3])
    elif len(t) == 2:
        t[0].append(t[1])

def p_options(t):
    '''options : options SPACE option
               | option'''
    if t[0] is None:
        t[0] = {}
    if len(t) == 4:
        t[0].update(t[1])
        t[0].update(t[3])
    elif len(t) == 2:
        t[0].update(t[1])

def p_option(t):
    '''option : ACTION EQUAL ACTION
              | ACTION EQUAL OPTIONVAL
              | ACTION EQUAL OPTIONVALS
              | ACTION EQUAL OPTIONVALD'''
    if t[0] is None:
        t[0] = {}
    if t[3][0] is "'":
        t[0][t[1]] = t[3].strip("'")
    elif t[3][0] is '"':
        t[0][t[1]] = t[3].strip('"')
    else:
        t[0][t[1]] = t[3]

def p_error(t):
    global parser_category
    global parser_action
    global parser_options
    global parser_error
    parser_category = None
    parser_action = None
    parser_options = None
    parser_error = None
    token = parser.token()
    if t is not None and token is not None:
        parser_error = "Syntax error at character: '%s', column: %d" % (str(token.value), int(token.lexpos))
    else:
        parser_error = "Syntax error at EOF"

parser = yacc.yacc()


from flask import Flask, Response, request
from flask_cors import CORS, cross_origin
import jwt
import os
from functools import wraps
import hashlib
import bandicoot.cli.api
import json
import datetime
import re
from ldap3 import Server, Connection
from ldap3.core.exceptions import LDAPSocketOpenError


app = Flask(__name__)
CORS(app)
app.secret_key = os.urandom(24)


def rest_request_is_valid(indata):
    """Validate Request"""
    # Verify required dict objects were present
    if indata is None or "options" not in indata or "category" not in indata or "action" not in indata:
        return False
    # Verify type of each
    if not (indata["options"] is None or isinstance(indata["options"], dict)) or not isinstance(indata["category"], basestring) or not isinstance(indata["action"], basestring):
        return False
    # Check options
    if isinstance(indata["options"], dict):
        for option in indata["options"]:
            # Check Key
            if not re.match(r'^[/_a-zA-Z0-9\-]+$', str(option)):
                return False
            # Check Value
            if not re.match(r'^[/_a-zA-Z0-9\*:\.\-=\?\~]+$', str(indata["options"][str(option)])):
                return False
    if not re.match(r'^[a-zA-Z0-9_\-/]+$', indata["category"]):
        return False
    if not re.match(r'^[a-zA-Z0-9_\-]+$', indata["action"]):
        return False

    return True


def create_token(user):
    global app
    payload = {
        # subject
        'sub': user,
        #issued at
        'iat': datetime.datetime.utcnow(),
        #expiry
        'exp': datetime.datetime.utcnow() + datetime.timedelta(days=1)
    }
 
    token = jwt.encode(payload, app.secret_key, algorithm='HS256')
    return token.decode('unicode_escape')
 

def parse_token(token):
    global app
    return jwt.decode(token, app.secret_key, algorithms='HS256')


def check_auth(username, password):
    """This function is called to check if a username /
    password combination is valid.
    """
    valid_auth = False

    # Hash of password provided
    m = hashlib.md5()
    m.update(password)
    password_md5 = m.hexdigest()

    # Check If Local User Can Be Authenticated
    post = bandicoot.cli.api.db.users.find_one({"username": username})
    if post is not None:
        if "password_md5" in post and post["password_md5"] == password_md5:
            valid_auth = True

    # LDAP Auth, if Local User Not Authenticated
    if valid_auth == False and bandicoot.cli.api.ldap_server is not None and bandicoot.cli.api.ldap_user_cn is not None:
        try:
            server = Server(bandicoot.cli.api.ldap_server, use_ssl=bandicoot.cli.api.ldap_use_ssl)
            conn = Connection(server, "uid=%s, %s" % (username, bandicoot.cli.api.ldap_user_cn), password)
            bind_success = conn.bind()
            if  bind_success == True:
                valid_auth = True
        except LDAPSocketOpenError:
            print("Failed to connect to LDAP server %s" % bandicoot.cli.api.ldap_server)

    return valid_auth


def authenticate():
    """Sends a 401 response that enables basic auth"""
    return Response(
    'Could not verify your access level for that URL.\n'
    'You have to login with proper credentials', 401,
    {'WWW-Authenticate': 'Basic realm="Login Required"'})


def requires_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        g = f.func_globals
        auth = request.authorization
        data = request.get_json(force=True)
        username = None
        password = None

        if auth is None:
            # user post username/password
            username = data["username"]
            password = data["password"]
        else:
            # username/password from Authorized Headers
            username = auth.username
            password = auth.password

        # Set username for decorated func
        g["username"] = username

        if not check_auth(username, password):
            return authenticate()
        return f(*args, **kwargs)
    return decorated


def token_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        g = f.func_globals

        if not request.headers.get('Authorization'):
            return Response(response="Missing authorization header", status=401)
        try:
            payload = parse_token(request.headers.get('Authorization').split()[1])
        except jwt.DecodeError:
            return Response(response="Token is invalid", status=401)
        except jwt.ExpiredSignature:
            return Response(response="Token has expired", status=401)

        # Set username for decorated func
        g["username"] = payload['sub']

        return f(*args, **kwargs)
    return decorated_function


@app.route("/login", methods=["POST"])
@requires_auth
def bandicoot_login():
    dat = None
    status = 200

    # Authenticate user, return a token
    dat = json.dumps({ "token": create_token(username) })

    # http response
    resp = Response(response=dat, status=status, mimetype="application/json")
    return(resp)


# Support Authorization headers for auth (bandicoot-cli uses this API)
@app.route("/", methods=["POST"])
@requires_auth
def bandicoot_base():
    indata = request.get_json()
    dat = None
    status = 200

    # Error on invalid request
    if not rest_request_is_valid(indata):
        return Response(response=json.dumps({"response": "  invalid request"}), status=400, mimetype="application/json")

    # Encrypt indata values that are sensitive
    bandicoot.cli.api.encrypt_dict(indata["options"])

    # Run Action
    dat = bandicoot.cli.api.parse_action(username, indata["category"], indata["action"], indata["options"])
    if dat is None:
        dat = json.dumps({"response": "  action not found"})

    # Audit Logging / History
    bandicoot.cli.api.log_action(username, {"result": dat, "category": indata["category"], "action": indata["action"], "options": indata["options"]})

    # http response
    resp = Response(response=dat, status=status, mimetype="application/json")
    return(resp)


# Support Token for auth (bandicoot-gui uses this API)
@app.route("/api", methods=["POST"])
@token_required
def bandicoot_api():
    indata = request.get_json()
    dat = None
    status = 200

    # Error on invalid request
    if not rest_request_is_valid(indata):
        return Response(response=json.dumps({"response": "  invalid request"}), status=400, mimetype="application/json")

    # Encrypt indata values that are sensitive
    bandicoot.cli.api.encrypt_dict(indata["options"])

    # Run Action
    dat = bandicoot.cli.api.parse_action(username, indata["category"], indata["action"], indata["options"])
    if dat is None:
        dat = json.dumps({"response": "  action not found"})
        status = 400

    # Audit Logging / History
    bandicoot.cli.api.log_action(username, {"result": dat, "category": indata["category"], "action": indata["action"], "options": indata["options"]})

    # http response
    resp = Response(response=dat, status=status, mimetype="application/json")
    return(resp)



import bandicoot.cli.api
from bandicoot.exceptions import DecryptWrongKeyException, DecryptNotClearTextException, DecryptException
import json
import subprocess
import hashlib
import datetime
import re
import shutil
import time
import multiprocessing
import threading
import sys
import Queue
import os


job_queue = {}
EOF = -1


def queue_support():
    def wrap(f):
        def wrapped_f(*args):
            global job_queue
            user = args[0]
            action = args[1]
            options = args[2]
            exit_event = multiprocessing.Event()
            q = multiprocessing.Queue()
            p = multiprocessing.Process(target=f, args=args+(exit_event,q))
            job_id = int(bandicoot.cli.api.counters_db_getNextSequence("jobid"))
            bandicoot.cli.api.db.jobs.insert_one({"_id": int(job_id), "start_time": time.time(), "user": user, "action": action, "options": options, "running": True, "response": ""})
            job_queue[job_id] = { "queue": q, "process": p, "exit_event": exit_event}
            p.start()
            return json.dumps({"queue_id": job_id})
        wrapped_f._original = f
        return wrapped_f
    return wrap


def options_validator(option_list, regexp):
    def wrap(f):
        def wrapped_f(*args):
            user = args[0]
            action = args[1]
            options = args[2]
            if options is not None:
                for key in options:
                    if key in option_list and not re.match(regexp, options[key]):
                        return json.dumps({"exit_code": 1, "response": "  option %s=%s has invalid characters" % (key, options[key])})
            return f(*args)
        return wrapped_f
    return wrap


def options_required(option_list):
    def wrap(f):
        def wrapped_f(*args):
            user = args[0]
            action = args[1]
            options = args[2]
            if options is not None:
                # Loop through required options
                for key in option_list:
                    # Check if each required option was given by the users options
                    if key not in options:
                        return json.dumps({"exit_code": 1, "response": "  %s option is required" % key})
            return f(*args)
        return wrapped_f
    return wrap


def options_supported(option_list):
    def wrap(f):
        def wrapped_f(*args):
            user = args[0]
            action = args[1]
            options = args[2]
            if options is not None:
                # Loop through supported options
                for key in options:
                    # Check if unsupported option was provided
                    if key not in option_list:
                        return json.dumps({"exit_code": 1, "response": "  %s option is not supported. Supported options are: %s." % (key, ", ".join(option_list))})
            return f(*args)
        return wrapped_f
    return wrap


def category_fix(options):
    if "category" in options:
        if options["category"] != "/":
            options["category"] = options["category"].rstrip("/")
            if len(options["category"]) >= 1:
                if options["category"][0] != "/":
                    options["category"] = "/" + options["category"]


def plugin_help(user, action, options):
    cursor = bandicoot.cli.api.db.actions.find()
    response = ""
    api_response = []
    compact_actions = {}
    dbaction_count = 0

    # Build Compact, Sortable, Dictinary like: {"roles": {"actions": ["list", "edit"], "num": 0}}
    for dbaction in bandicoot.cli.api.builtin_actions + list(cursor):
        if not bandicoot.cli.api.roles_has_permission(user, {"category": dbaction["category"], "action": dbaction["action"]}, {}):
            continue

        if dbaction["category"] not in compact_actions:
            compact_actions[dbaction["category"]] = {"actions" : [dbaction["action"]], "descs": [dbaction["desc"]]}
            compact_actions[dbaction["category"]]["num"] = dbaction_count
        else:
            compact_actions[dbaction["category"]]["actions"].append(dbaction["action"])
            compact_actions[dbaction["category"]]["descs"].append(dbaction["desc"])
            compact_actions[dbaction["category"]]["num"] = dbaction_count
        dbaction_count += 1

    # Print Compact Help for the top level
    for db_item in sorted(compact_actions.items(), key=lambda x: x[1]['num']):
        dbaction["category"] = db_item[0]
        dbaction["actions"] = db_item[1]["actions"]
        dbaction["descs"] = db_item[1]["descs"]
        # Help specifying command
        if action["category"] == "/help":
            a_count = 0
            for actionline in dbaction["actions"]:
                if dbaction["category"].strip("/") == action["action"].strip("/"):
                    response += "  %s\t%s\n" % (actionline, dbaction["descs"][a_count])
                    a_count += 1
            api_response.append({"category": "", "action": dbaction["actions"], "desc": dbaction["descs"]})
        # Help by itself, no command
        else:
            category_str = dbaction['category'].strip("/").replace("/", " ")
            if category_str is None or len(category_str) <= 0:
                api_response.append({"category": "", "action": dbaction["actions"], "desc": dbaction["descs"]})
                for actionline in dbaction["actions"]:
                    response += "  %s\n" % actionline
            else:
                api_response.append({"category": dbaction["category"].strip("/").replace("/", " "), "action": dbaction["actions"], "desc": dbaction["descs"]})
                response += "  %s [%s]\n" % (dbaction["category"].strip("/").replace("/", " "), "|".join(dbaction["actions"]))

    # Append the exit builtin implemented on the client side
    response += "  exit \t\t\n"

    return json.dumps({"exit_code": 0, "response": response, "api_response": api_response})


def plugin_ping(user, action, options):
    return json.dumps({"exit_code": 0, "response": "  pong"})


@options_supported(option_list=["username", "password"])
@options_required(option_list=["username", "password"])
@options_validator(option_list=["username"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_users_add(user, action, options):
    result = bandicoot.cli.api.db.users.find_one({"username": options["username"]})
    if result is None:
        m = hashlib.md5()
        m.update(str(options["password"]))
        password_md5 = str(m.hexdigest())
        post = {"username": options["username"], "password_md5": password_md5}
        bandicoot.cli.api.db.users.insert_one(post)
        return json.dumps({"exit_code": 0, "response": "  created user %s" % options["username"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  user %s already exists" % options["username"]})


@options_supported(option_list=["username"])
@options_required(option_list=["username"])
@options_validator(option_list=["username"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_users_del(user, action, options):
    post = {"username": options["username"]}
    result = bandicoot.cli.api.db.users.delete_many(post)
    if result.deleted_count > 0:
        return json.dumps({"exit_code": 0, "response": "  deleted user %s" % options["username"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  user %s does not exist" % options["username"]})


@options_supported(option_list=["username", "password"])
@options_required(option_list=["password"])
@options_validator(option_list=["username"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_users_edit(user, action, options):
    # Hash the password
    m = hashlib.md5()
    m.update(options["password"])
    password_md5 = str(m.hexdigest())

    # If no username was specified, by default edit the current user
    if "username" not in options:
        options["username"] = user

    result = bandicoot.cli.api.db.users.update_one({"username": options["username"]},
            {"$set": {"password_md5": password_md5},})
    if result.matched_count > 0:
        return json.dumps({"exit_code": 0, "response": "  modified user %s" % options["username"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  user %s does not exist" % options["username"]})


def plugin_users_list(user, action, options):
    result = ""
    cursor = bandicoot.cli.api.db.users.find()
    for doc in list(cursor):
        result += "  %s\n" % doc["username"]
    return json.dumps({"exit_code": 0, "response": result.rstrip()}) # Do not return the last character (carrage return)


@options_required(option_list=["name", "category", "action", "plugin", "desc"])
@options_validator(option_list=["name", "plugin", "action"], regexp=r'^[a-zA-Z0-9_\-]+$')
@options_validator(option_list=["category"], regexp=r'^[a-zA-Z0-9_\-/]+$')
def plugin_actions_add(user, action, options):
    dat = None

    category_fix(options)

    find_result = bandicoot.cli.api.db.actions.find_one({"name": options["name"]})
    if find_result is None:
        result = bandicoot.cli.api.db.actions.insert_one(options)
        dat = json.dumps({"exit_code": 0, "response": "  created action %s" % options["name"]})
    else:
        dat = json.dumps({"exit_code": 1, "response": "  action %s already exists" % options["name"]})
    return dat


def plugin_command(user, action, options):
    result = ""
    if "command_run" not in action:
        return json.dumps({"exit_code": 1, "response": "  command_run required in action"})
    cmd = action["command_run"].split()
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    for line in p.stdout:
        result += "  %s\n" % line
    p.wait()
    result += "  return code: %d\n"  % p.returncode
    return json.dumps({ "exit_code": 0, "response": result})


@options_required(option_list=["name"])
@options_validator(option_list=["name", "plugin", "action"], regexp=r'^[a-zA-Z0-9_\-]+$')
@options_validator(option_list=["category"], regexp=r'^[a-zA-Z0-9_\-/]+$')
def plugin_actions_edit(user, action, options):
    category_fix(options)

    result = bandicoot.cli.api.db.actions.update_one({"name": options["name"]},
            {"$set": options})
    if result.matched_count > 0:
        return json.dumps({"exit_code": 0, "response": "  modified action %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  action %s does not exist" % options["name"]})


@options_supported(option_list=["name"])
@options_required(option_list=["name"])
@options_validator(option_list=["name"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_actions_del(user, action, options):
    dat = None

    post = {"name": options["name"]}
    result = bandicoot.cli.api.db.actions.delete_many(post)
    if result.deleted_count > 0:
        dat = json.dumps({"exit_code": 0, "response": "  deleted action %s" % options["name"]})
    else:
        dat = json.dumps({"exit_code": 1, "response": "  action %s does not exist" % options["name"]})
    return dat


def plugin_actions_list(user, action, options):
    result = ""
    cursor = bandicoot.cli.api.db.actions.find()
    for doc in list(cursor):
        for key in sorted(doc):
            if key not in ["_id"]:
                result += '  %s="%s" ' % (key, doc[key])
        result += "\n"
    return json.dumps({"exit_code": 0, "response": result.rstrip()}) # Do not return the last character (carrage return)


@options_required(option_list=["name"])
@options_validator(option_list=["name"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_roles_add(user, action, options):
    result = bandicoot.cli.api.db.roles.find_one({"name": options["name"]})
    if result is None:
        post = options
        bandicoot.cli.api.db.roles.insert_one(post)
        return json.dumps({"exit_code": 0, "response": "  created role %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  role %s already exists" % options["name"]})


@options_required(option_list=["name"])
@options_validator(option_list=["name"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_roles_edit(user, action, options):
    result = bandicoot.cli.api.db.roles.update_one({"name": options["name"]},
            {"$set": options})
    if result.matched_count > 0:
        return json.dumps({"exit_code": 0, "response": "  modified role %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  role %s does not exist" % options["name"]})


@options_supported(option_list=["name"])
@options_required(option_list=["name"])
@options_validator(option_list=["name"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_roles_del(user, action, options):
    post = {"name": options["name"]}
    result = bandicoot.cli.api.db.roles.delete_many(post)
    if result.deleted_count > 0:
        return json.dumps({"exit_code": 0, "response": "  deleted role %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  role %s does not exist" % options["name"]})


def plugin_roles_list(user, action, options):
    result = ""
    cursor = bandicoot.cli.api.db.roles.find()
    for doc in list(cursor):
        for key in sorted(doc):
            if key not in ["_id"]:
                result += '  %s="%s" ' % (key, doc[key])
        result += "\n"
    return json.dumps({"exit_code": 0, "response": result.rstrip()}) # Do not return the last character (carrage return)


@options_required(option_list=["name"])
@options_validator(option_list=["name"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_secrets_add(user, action, options):
    result = bandicoot.cli.api.db.secrets.find_one({"name": options["name"]})
    if result is None:
        post = options
        bandicoot.cli.api.db.secrets.insert_one(post)
        return json.dumps({"exit_code": 0, "response": "  created secret %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  secret %s already exists" % options["name"]})


@options_required(option_list=["name"])
@options_validator(option_list=["name"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_secrets_edit(user, action, options):
    result = bandicoot.cli.api.db.secrets.update_one({"name": options["name"]},
            {"$set": options})
    if result.matched_count > 0:
        return json.dumps({"exit_code": 0, "response": "  modified secret %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  secret %s does not exist" % options["name"]})


@options_supported(option_list=["name"])
@options_required(option_list=["name"])
@options_validator(option_list=["name"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_secrets_del(user, action, options):
    post = {"name": options["name"]}
    result = bandicoot.cli.api.db.secrets.delete_many(post)
    if result.deleted_count > 0:
        return json.dumps({"exit_code": 0, "response": "  deleted secret %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  secret %s does not exist" % options["name"]})


@options_supported(option_list=["oldpw"])
def plugin_secrets_encryptpw(user, action, options):
    def update_secret_pw(name, options):
        result = bandicoot.cli.api.db.secrets.update_one({"name": name},
                {"$set": options})
        if result.matched_count > 0:
            return True
        else:
            return False
    result = ""
    cursor = bandicoot.cli.api.db.secrets.find()
    # secrets encryptpw --- encrypt clear text pw using the current password
    # secrets encryptpw oldpw=XXXX --- re-encrypt secrets that use oldpw to use the current password
    for doc in list(cursor):
        if "secret" in doc:
            # Detect status of secret
            decrypted_secret = ""
            encrypted_secret = None
            what_to_do = "nothing"
            old_encrypt_password = None # Default assumes secret is clear text, if no oldpw is provided this is assumed
            new_encrypt_password = None # use global pw
            if options is not None and "oldpw" in options:
                # Old password provided, so use it
                old_encrypt_password = str(options["oldpw"])


            # Detect The Problem With Decryption
            try:
                decrypted_secret = bandicoot.cli.api.decrypt_str(doc["secret"], encrypt_password=new_encrypt_password)
            except DecryptWrongKeyException:
                # P
                what_to_do = "updatepw"
            except DecryptNotClearTextException:
                # User said it was clear text, but its not. Skip, Do Nothing.
                what_to_do = "notcleartext"
            # Prefix is "__bandicoot_encrypted__:"
            if decrypted_secret == doc["secret"][len("__bandicoot_encrypted__:"):]:
                # Clear text to encrypted
                what_to_do = "encrypt"

            if what_to_do == "updatepw":
                try:
                    decrypted_secret = bandicoot.cli.api.decrypt_str(doc["secret"], encrypt_password=old_encrypt_password)
                except DecryptException:
                    result += "secret %s failed to update to new password\n" % doc["name"]
                    continue
                encrypted_secret = bandicoot.cli.api.encrypt_str(decrypted_secret, new_encrypt_password)
                result += "secret %s updated to new password\n" % doc["name"]
            elif what_to_do == "encrypt":
                try:
                    decrypted_secret = bandicoot.cli.api.decrypt_str(doc["secret"], encrypt_password=old_encrypt_password)
                except DecryptException:
                    result += "secret %s failed to update because the secret is not clear text\n" % doc["name"]
                    continue
                encrypted_secret = bandicoot.cli.api.encrypt_str(decrypted_secret, new_encrypt_password)
                result += "secret %s encrypted using new password\n" % doc["name"]

            # Secret Was Updated
            if encrypted_secret is not None:
                update_secret_pw(doc["name"], {"secret": encrypted_secret})

    return json.dumps({"exit_code": 0, "response": result})


def plugin_secrets_list(user, action, options):
    result = ""
    cursor = bandicoot.cli.api.db.secrets.find()
    for doc in list(cursor):
        if "secret" in doc:
            # Detect status of secret
            decrypted_secret = ""
            doc["status"] = "encrypted"
            try:
                decrypted_secret = bandicoot.cli.api.decrypt_str(doc["secret"])
            except DecryptWrongKeyException:
                doc["status"] = "wrongpw"
            except DecryptNotClearTextException:
                doc["status"] = "noencryptpw"
            # Prefix is "__bandicoot_encrypted__:"
            if decrypted_secret == doc["secret"][len("__bandicoot_encrypted__:"):]:
                doc["status"] = "cleartext"

            # Do Not Print Encrypted Secret
            doc["secret"] = "..."
        for key in sorted(doc):
            if key not in ["_id"]:
                result += '  %s="%s" ' % (key, doc[key])
        result += "\n"
    return json.dumps({"exit_code": 0, "response": result.rstrip()}) # Do not return the last character (carrage return)


def plugin_plugins_list(user, action, options):
    return json.dumps({"exit_code": 0, "response": "\n  ".join(bandicoot.cli.api.plugins.keys())})


def plugin_logs(user, action, options):
    result = ""

    # Index is required because of the sorting
    # Not sure if this should be done at each call or only once when the first log is created
    bandicoot.cli.api.db.logs.create_index([("date", 1)])
    bandicoot.cli.api.db.inventory.changes.create_index([("date", 1)])

    if options is not None and ("name" in options or ("type" in options and options["type"] == "changes")):
        # List changes/logs for a specific inventory host
        result += "  inventory_item\t\tdesc\t\tjob_id\t\tdate\n"
        if "name" in options:
            # Show changes for a specific inventory item
            cursor = bandicoot.cli.api.db.inventory.changes.find({"name": options["name"]}).sort("date", 1)
        else:
            # Show all changes
            cursor = bandicoot.cli.api.db.inventory.changes.find().sort("date", 1)
        for doc in list(cursor):
            if "date" in doc: # Backward compat
                result += '  %s\t"%s"\t%s\t%s\n' % (doc["name"], doc["desc"], doc["job_id"], "{:%m/%d/%Y %M:%H}".format(doc["date"]))
    else: # type=requests is the default
        # Default List all requests to api server
        result += "  category\t\taction\t\toptions\t\tdate\n"
        cursor = bandicoot.cli.api.db.logs.find().sort("date", 1)
        for doc in list(cursor):
            # Backward compat when user field did not exist
            if "user" not in doc:
                doc["user"] = "unknown"
            # Backward compat when result field did not exist
            if "result" not in doc:
                doc["result"] = "unknown"
            # Backward compat when date field did not exist
            if "date" not in doc:
                doc["date"] = datetime.date(1970, 1, 1) # unknown
            result += "  %s\t%s\t%s\t%s\t%s\n" % (doc["user"], doc["category"], doc["action"], doc["options"], "{:%m/%d/%Y %M:%H}".format(doc["date"]))
    return json.dumps({"exit_code": 0, "response": result})


@queue_support()
def plugin_ansible(user, action, options, exit_event, q):
    def process_stdout(p, q):
        for line in p.stderr:
            q.put("  %s\n" % line)
        for line in p.stdout:
            q.put("  %s\n" % line)
        p.wait()
        #time.sleep(0.1) # Delay For Queue Operations
        return 0

    ansible_options = ""
    temp_location = "/tmp/bandicoot/%s" % str(time.time())

    # Required options to be included in action
    for option in ["source_url", "playbook"]:
        if option not in action:
            return json.dumps({"exit_code": 1, "response": "  %s required in action" % option})

    # Sudo
    if "sudo" in action and action["sudo"] == "yes":
        ansible_options += "-s "

    # Git
    cmd = str("git clone %s %s" % (action["source_url"], temp_location)).split()
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    thread = threading.Thread(target=process_stdout, args=(p,q))
    thread.start()
    while thread.isAlive() and not exit_event.is_set():
        time.sleep(0.25)

    if thread.isAlive():
        p.kill()

    time.sleep(0.25)
    thread.join()

    # Ansible
    cmd = str("ansible-playbook %s %s" % (ansible_options, action["playbook"])).split()
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=temp_location)
    thread = threading.Thread(target=process_stdout, args=(p,q))
    thread.start()
    while thread.isAlive() and not exit_event.is_set():
        time.sleep(0.25)

    if thread.isAlive():
        p.kill()

    time.sleep(0.25)
    thread.join()

    # Delete temporary git directory
    if os.path.isdir(temp_location):
        shutil.rmtree(temp_location)

    q.put(EOF)
    time.sleep(0.1) # Delay For Queue
    #sys.exit(0)
    return json.dumps({"exit_code": 0, "response": "  success"}) # For unittesting


@options_supported(option_list=["id"])
@options_required(option_list=["id"])
@options_validator(option_list=["id"], regexp=r'^[0-9]+$')
def plugin_jobs_status(user, action, options):
    global job_queue
    exit_code = 0

    result = bandicoot.cli.api.db.jobs.find_one({"_id": int(options["id"])})
    if result is None:
        exit_code = 1
        return json.dumps({"exit_code": 1, "response": "  The job id %s does not match a job" % str(options["id"]), "exit_code": exit_code})
    else:
        int_id = int(options["id"])
        if result["user"] != user:
            exit_code = 1
            return json.dumps({"exit_code": 1, "response": "  The job %s, is owned by another user" % str(int_id), "exit_code": exit_code})

        # Get all items from queue until its empty or EOF is reached
        while True:
            try:
                if result["_id"] not in job_queue:
                    result["running"] = False
                    bandicoot.cli.api.db.jobs.update_one({"_id": result["_id"]}, {"$set": {"running": result["running"]},})
                    # Job already ran, just return the result
                    break

                qitem = job_queue[result["_id"]]["queue"].get_nowait()
                if qitem != EOF:
                    # New Data from job!
                    result["response"] += qitem
                    bandicoot.cli.api.db.jobs.update_one({"_id": result["_id"]}, {"$set": {"response": result["response"]},})
                else:
                    # EOF, job is finished
                    result["running"] = False
                    bandicoot.cli.api.db.jobs.update_one({"_id": result["_id"]}, {"$set": {"running": result["running"]},})
                    result["end_time"] = time.time()
                    bandicoot.cli.api.db.jobs.update_one({"_id": result["_id"]}, {"$set": {"end_time": result["end_time"]},})

                    # Discover New Inventory
                    tmp_hosts = {}
                    current_task_name = "unknown"

                    for line in result["response"].split("\n"):
                        line = line.strip()
                        line = line.strip("\n")
                        # Match Task name, example:   TASK [setup] *******************************************************************
                        m = re.match(r'^\s*TASK\s+\[(.*?)\]', line)
                        if m:
                            current_task_name = m.group(1)
                        # Match changed|ok|skipped|fatal, example:   ok: [ec2-52-89-49-1.us-west-2.compute.amazonaws.com]
                        m = re.match(r'^\s*([a-z]+):\s+\[(.*?)\]', line)
                        if m:
                            change_state = m.group(1)
                            hostname = m.group(2)
                            if hostname not in tmp_hosts:
                                tmp_hosts[hostname] = []
                            if change_state == "changed" or change_state == "fatal": # Only log changes or failed changes
                                tmp_hosts[hostname].append({"name": hostname, "date": datetime.datetime.utcnow(), "desc": current_task_name, "job_id": result["_id"]})

                    for hostname in tmp_hosts:
                        dbresult = bandicoot.cli.api.db.inventory.hosts.find_one({"name": hostname})
                        if dbresult is None:
                            # New Inventory Item discovered
                            bandicoot.cli.api.db.inventory.hosts.insert_one({"name": hostname})

                        # Log Inventory Changelog
                        for document in tmp_hosts[hostname]:
                            bandicoot.cli.api.db.inventory.changes.insert_one(document)

                    break
            except Queue.Empty:
                break

        return json.dumps({"response": result["response"], "finished": not result["running"], "exit_code": exit_code})


def plugin_jobs_list(user, action, options):
    result = "  Job ID\tIs Running?\tUser\tCommand\n"
    api_result = []
    cursor = bandicoot.cli.api.db.jobs.find()
    for doc in list(cursor):
        is_running = doc["_id"] in job_queue and doc["running"]
        api_result.append({"_id": doc["_id"], "is_running": is_running, "user": doc["user"], "category": doc["action"]["category"], "action": doc["action"]["action"]})
        result += "  %s\t\t%s\t\t%s\t\t%s/%s\n" % (str(doc["_id"]), str(is_running),
                                              str(doc["user"]),
                                              str(doc["action"]["category"]).rstrip("/"),
                                              str(doc["action"]["action"]))

    return json.dumps({"exit_code": 0, "response": result, "api_response": api_result})


@options_supported(option_list=["id"])
@options_required(option_list=["id"])
@options_validator(option_list=["id"], regexp=r'^[0-9]+$')
def plugin_jobs_kill(user, action, options):
    global job_queue
    
    result = bandicoot.cli.api.db.jobs.find_one({"_id": int(options["id"])})
    if result is None:
        return json.dumps({"exit_code": 1, "response": "  The job id %s does not match a job" % str(options["id"])})
    else:
        int_id = int(options["id"])
        if result["running"] == False:
            return json.dumps({"exit_code": 1, "response": "  The job %s, was already terminated" % str(int_id)})
        elif result["user"] != user:
            return json.dumps({"exit_code": 1, "response": "  The job %s, is owned by another user" % str(int_id)})
        job_queue[int_id]["exit_event"].set() # Set event to trigger exit
        job_queue[int_id]["process"].join() # Wait for Plugin to exit
        result["running"] = False
        bandicoot.cli.api.db.jobs.update_one({"_id": result["_id"]}, {"$set": {"running": result["running"]},})
        return json.dumps({"exit_code": 0, "response": "  The job %s, was terminated" % str(int_id)})


@options_supported(option_list=["name", "user", "category", "action", "minute", "hour", "day_of_month", "month", "day_of_week"])
@options_required(option_list=["name", "category", "action"])
@options_validator(option_list=["name", "user", "action"], regexp=r'^[a-zA-Z0-9_\-]+$')
@options_validator(option_list=["category"], regexp=r'^[/a-zA-Z0-9_\-]+$')
@options_validator(option_list=["minute", "hour", "day_of_month", "month", "day_of_week"], regexp=r'^([0-9]+|\*)$')
def plugin_schedules_add(user, action, options):
    result = bandicoot.cli.api.db.schedules.find_one({"name": options["name"]})
    if result is None:
        post = options
        # Prevent Users from setting up crons for other users
        if "user" in post and post["user"] != user:
            return json.dumps({"exit_code": 1, "response": "  You cannot set the cron user to anyone but your username %s." % user})
        # Always set the cron by default to run as the user
        if "user" not in post:
            post["user"] = user

        bandicoot.cli.api.db.schedules.insert_one(post)
        return json.dumps({"exit_code": 0, "response": "  created schedule %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  schedule %s already exists" % options["name"]})


@options_supported(option_list=["name", "user", "category", "action", "minute", "hour", "day_of_month", "month", "day_of_week"])
@options_required(option_list=["name"])
@options_validator(option_list=["name", "user", "action"], regexp=r'^[a-zA-Z0-9_\-]+$')
@options_validator(option_list=["category"], regexp=r'^[/a-zA-Z0-9_\-]+$')
@options_validator(option_list=["minute", "hour", "day_of_month", "month", "day_of_week"], regexp=r'^([0-9]+|\*)$')
def plugin_schedules_edit(user, action, options):
    # Prevent Users from setting up crons for other users
    if "user" in options and options["user"] != user:
        return json.dumps({"exit_code": 1, "response": "  You cannot set the cron user to anyone but your username %s." % user})

    result = bandicoot.cli.api.db.schedules.update_one({"name": options["name"]},
            {"$set": options})
    if result.matched_count > 0:
        return json.dumps({"exit_code": 0, "response": "  modified schedule %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  schedule %s does not exist" % options["name"]})


@options_supported(option_list=["name"])
@options_required(option_list=["name"])
@options_validator(option_list=["name"], regexp=r'^[a-zA-Z0-9_\-]+$')
def plugin_schedules_del(user, action, options):
    post = {"name": options["name"]}
    result = bandicoot.cli.api.db.schedules.delete_many(post)
    if result.deleted_count > 0:
        return json.dumps({"exit_code": 0, "response": "  deleted schedule %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  schedule %s does not exist" % options["name"]})


def plugin_schedules_list(user, action, options):
    result = ""
    cursor = bandicoot.cli.api.db.schedules.find()
    for doc in list(cursor):
        for key in sorted(doc):
            if key not in ["_id"]:
                result += '  %s="%s" ' % (key, doc[key])
        result += "\n"
    return json.dumps({"exit_code": 0, "response": result.rstrip()}) # Do not return the last character (carrage return)


def plugin_inventory_list(user, action, options):
    result = ""
    cursor = bandicoot.cli.api.db.inventory.hosts.find()
    for doc in list(cursor):
        result += "  %s\n" % doc["name"]
    return json.dumps({"exit_code": 0, "response": result.rstrip()}) # Do not return the last character (carrage return)


@options_supported(option_list=["name"])
@options_required(option_list=["name"])
def plugin_inventory_del(user, action, options):
    post = {"name": options["name"]}
    result = bandicoot.cli.api.db.inventory.hosts.delete_many(post)
    if result.deleted_count > 0:
        return json.dumps({"exit_code": 0, "response": "  deleted inventory item %s" % options["name"]})
    else:
        return json.dumps({"exit_code": 1, "response": "  inventory item %s does not exist" % options["name"]})


@options_supported(option_list=["type"])
def plugin_stats(user, action, options):
    from ascii_graph import Pyasciigraph
    response = ""
    stat_data = []
    stat_type = "users"
    stat_title = ""
    stat_type_supported = ("users", "system", "jobs")
    if options is not None and "type" in options and options["type"] in stat_type_supported:
        stat_type = options["type"]

    # Sum up Job By User
    if stat_type == "users":
        stat_title = "Jobs Submitted Per User"
        user_stats = {}
        cursor = bandicoot.cli.api.db.jobs.find()
        for doc in list(cursor):
            if doc["user"] not in user_stats:
                user_stats[doc["user"]] = 1
            else:
                user_stats[doc["user"]] += 1
        stat_data = [(k, v) for k, v in user_stats.iteritems()] # Convert Dict to List of Tuples
    elif stat_type == "system":
        stat_title = "Changes Per Inventory Item"
        system_stats = {}
        cursor = bandicoot.cli.api.db.inventory.changes.find().sort("date", 1)
        for doc in list(cursor):
            if doc["name"] not in system_stats:
                system_stats[doc["name"]] = 1
            else:
                system_stats[doc["name"]] += 1
        stat_data = [(k, v) for k, v in system_stats.iteritems()] # Convert Dict to List of Tuples
    elif stat_type == "jobs":
        stat_title = "Jobs Submitted By Date"
        job_stats = {}
        cursor = bandicoot.cli.api.db.inventory.changes.find().sort("date", 1)
        for doc in list(cursor):
            if "date" in doc:
                doc_date = "{:%m/%d/%Y}".format(doc["date"])
                if doc_date not in job_stats:
                    job_stats[doc_date] = 1
                else:
                    job_stats[doc_date] += 1
        stat_data = [(k, v) for k, v in job_stats.iteritems()] # Convert Dict to List of Tuples

    # Graph Data
    graph = Pyasciigraph()
    for line in graph.graph(stat_title, stat_data):
        response += "%s\n" % line

    return json.dumps({"exit_code": 0, "response": "  %s" % response})



""" Command Line Interface Module """
import optparse
import sys
import os
import requests
import json
import yaml
import getpass
import curses
import time
import signal
from bandicoot.parser import yacc

session = requests.Session()
sig_bg_pressed = 0
sig_kill_pressed = 0

# UNICODE
import locale
locale.setlocale(locale.LC_ALL, '')

# catch ctrl-z
def sig_background(signum, frame):
    global sig_bg_pressed
    sig_bg_pressed = 1


# catch ctrl-c
def sig_kill(signum, frame):
    global sig_kill_pressed
    sig_kill_pressed = 1


# signal for ctrl-z
signal.signal(signal.SIGTSTP, sig_background)
# signal for ctrl-c
signal.signal(signal.SIGINT, sig_kill)


class Cli(object):
    """ bandicoot CLI """

    def __init__(self):
        """ Setup Arguments and Options for CLI """
        # Parse CLI Arguments
        parser = optparse.OptionParser()
        parser.add_option("-u", "--user", dest="user",
                          help="bandicoot username",
                          metavar="USER",
                          default=None)
        parser.add_option("-s", "--server", dest="server",
                          help="IP address or hostname of bandicoot-api server",
                          metavar="SERVER",
                          default=None)
        parser.add_option("-p", "--port", dest="port",
                          help="tcp port of bandicoot-api server",
                          metavar="PORT",
                          default=None)
        parser.add_option("-t", "--insecure", dest="is_secure",
                          help="Do Not Use SSL",
                          metavar="SECURE",
                          action="store_false",
                          default=True)
        parser.add_option("-k", "--no-check-certificates", dest="is_ssl_verify",
                          help="Ignore Unverified Certificate",
                          metavar="VERIFY",
                          action="store_false",
                          default=True)
        # Assign values from cli
        (options, args) = parser.parse_args()
        self.user = options.user
        self.server = options.server
        self.port = options.port
        self.is_secure = options.is_secure
        self.is_ssl_verify = options.is_ssl_verify
        self.interactive_mode = True
        self.noninteractive_commands = []
        self.password = None
        self.app_running = True

        # Do Not Display SSL Verify Warning to stderr
        requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)

        # Non-Interactive Command Parsing
        if len(args) > 0:
            self.interactive_mode = False
            for command in args:
                self.noninteractive_commands.append(command)

        # Assign values from conf
        bandicoot_config_locations = [os.path.expanduser("~")+"/.bandicoot.conf", "/etc/bandicoot.conf"]
        bandicoot_conf_obj = {}
        for bandicoot_conf in bandicoot_config_locations:
            if os.path.isfile(bandicoot_conf):
                with open(bandicoot_conf, 'r') as stream:
                    try:
                        bandicoot_conf_obj = yaml.load(stream)
                    except yaml.YAMLError as excep:
                        print("%s\n" % excep)
        if self.user is None and "user" in bandicoot_conf_obj:
            self.user = str(bandicoot_conf_obj["user"])
        if self.password is None and "password" in bandicoot_conf_obj:
            self.password = str(bandicoot_conf_obj["password"])
        if self.server is None and "server" in bandicoot_conf_obj:
            self.server = str(bandicoot_conf_obj["server"])
        if self.port is None and "port" in bandicoot_conf_obj:
            self.port = int(bandicoot_conf_obj["port"])
        if self.is_secure == True and "secure" in bandicoot_conf_obj:
            self.is_secure = bool(bandicoot_conf_obj["secure"])
        if self.is_ssl_verify == True and "ssl_verify" in bandicoot_conf_obj:
            self.is_ssl_verify = bool(bandicoot_conf_obj["ssl_verify"])

        # Assign Default values if they were not specified at the cli or in the conf
        if self.user is None:
            self.user = "superadmin"
        if self.server is None:
            self.server = "127.0.0.1"
        if self.port is None:
            self.port = 8088

        self.url = "%s://%s:%d" % ("https" if self.is_secure else "http", str(self.server), int(self.port))
        self.screen = None
        self.history = []

    def welcome(self):
        """ Welcome Message """
        self.screen.addstr("======================\n")
        self.screen.addstr("Welcome To bandicoot\n")
        if "pong" in self.action_ping():
            self.screen.addstr("Connected to Server %s\n" % self.url)
        else:
            print("Failed connecting to server %s\n" % self.url)
            self.exit(1)
        self.screen.addstr("======================\n")

    def login_prompt(self):
        auth_success = False
        default_username = "superadmin"
        default_pw = "superadmin"

        if self.user is None:
            self.user = raw_input("Username: ")

        for trycount in [1, 2, 3]:
            if self.password is None:
                self.password = getpass.getpass()
            if "pong" in self.action_ping():
                auth_success = True
                if self.user == default_username and self.password == default_pw:
                    for change_trycount in [1, 2, 3]:
                        print("Changing Password From Default")
                        new_password = getpass.getpass("Enter New Password: ")
                        new_password_repeat = getpass.getpass("Enter New Password Again: ")
                        if new_password == new_password_repeat:
                            if self.action_changepw(self.user, new_password) is not "":
                                self.password = new_password
                                break
                break
            else:
                self.password = None

        if auth_success == False:
            print("Login Failed\n")
            self.exit(1)


    def exit(self, val):
        sys.exit(val)
        self.app_running = False # For unit testing

    def action_changepw(self, username, password):
        data = self.run_action(self.get_action_from_command("users edit username='%s' password='%s'"
            % (username, password)))
        if data is not None:
            return data["response"]
        else:
            return ""

    def action_ping(self):
        data = self.run_action(self.get_action_from_command("ping"))
        if data is not None:
            return data["response"]
        else:
            return ""

    def is_action_quit(self, action):
        return len(action) == 1 and ( action[0] == "quit" or action[0] == "exit" )

    def action_quit(self):
        self.screen.addstr("  Goodbye!\n")
        self.exit(0)

    def run_action(self, actionjson):
        r = session.post(self.url, verify=self.is_ssl_verify, headers={'Content-Type': 'application/json'},
            auth=(self.user, self.password), data=json.dumps(actionjson))

        if r.status_code == requests.codes.ok:
            return json.loads(r.text)
        else:
            return None

    def get_action_from_command(self, line):
        if line is not None and len(line) > 0:
            # Reset Parser Variables
            yacc.parser_category = None
            yacc.parser_action = None
            yacc.parser_options = None
            yacc.parser_error = None
            # Parse line input
            yacc.parser.parse(line)
            # Return Action Object
            return {'category': yacc.parser_category, "action": yacc.parser_action, "options": yacc.parser_options}
        else:
            return {'category': None, "action": None, "options": None}

    def startshell(self, arg):
        self.screen = curses.initscr()
        self.welcome()
        curses.curs_set(1)
        self.screen.addstr("bandicoot> ")
        self.screen.keypad(1)
        self.screen.scrollok(1)

        # left and right key
        cursor_offset = 0

        # ctrl-u
        history_index = 0

        # ctrl-r
        search_mode = False
        last_match = None

        line = ""
        while self.app_running:
            s = self.screen.getch()

            # Ascii
            if s >= 32 and s <= 126:
                if cursor_offset >= 0:
                    # cursor at end of line
                    line += chr(s)
                    self.screen.addstr(chr(s))
                elif cursor_offset <= len(line)*-1:
                    # cursor at beginning of line
                    line = chr(s) + line
                    self.screen.insstr(chr(s))
                    (y, x) = self.screen.getyx()
                    self.screen.move(y, x+1)
                else:
                    # cursor in the middle of the line
                    line = line[:len(line)+cursor_offset] + chr(s) + line[len(line)+cursor_offset:]
                    self.screen.insstr(chr(s))
                    (y, x) = self.screen.getyx()
                    self.screen.move(y, x+1)
                if search_mode:
                    match = None
                    for item in reversed(self.history):
                        if line in item:
                            match = item
                            break
                    if match is None:
                        self.screen.addstr(y, 0, "(reverse-i-search)`':")
                        self.screen.addstr(y, len("(reverse-i-search)`':"), line)
                        self.screen.clrtoeol()
                    else:
                        (y, x) = self.screen.getyx()
                        self.screen.addstr(y, 0, "(reverse-i-search)`':")
                        self.screen.addstr(y, len("(reverse-i-search)`':"), match)
                        self.screen.clrtoeol()
                        last_match = match
                history_index = 0
            # Finished With Line Input
            elif s == ord("\n"):
                (y, x) = self.screen.getyx()
                self.screen.move(y, len("bandicoot> ")+len(line))
                self.screen.addstr("\n")
                if search_mode:
                    if match is not None:
                        result = self.shell_parse_line(match)
                        self.screen.addstr(result)
                else:
                    result = self.shell_parse_line(line)
                    if result is not None:
                        self.screen.addstr(result.encode("UTF-8"))
                self.screen.addstr("\nbandicoot> ")
                line = ""
                history_index = 0
                cursor_offset = 0
                search_mode = False
            # Backspace
            elif s == curses.KEY_BACKSPACE or s == 127 or s == curses.erasechar():
                (y, x) = self.screen.getyx()
                if len(line) > 0 and x > len("bandicoot> "):
                    line = line[:len(line)+cursor_offset-1] + line[len(line)+cursor_offset:]
                    self.screen.delch(y, x-1)
                history_index = 0
            # Ctrl-u, clear line
            elif s == 21:
                (y, x) = self.screen.getyx()
                self.screen.addstr(y, 0, "bandicoot> ")
                self.screen.clrtoeol()
                line = ""
                history_index = 0
            # Ctrl-r, search
            elif s == 18:
                search_mode = True
                (y, x) = self.screen.getyx()
                self.screen.addstr(y, 0, "(reverse-i-search)`':")
                self.screen.clrtoeol()
                line = ""
                history_index = 0
            elif s == curses.KEY_UP:
                if len(self.history) < 1:
                    # prevent divide by zero when history is 0
                    continue
                history_index += 1
                cursor_offset = 0
                (y, x) = self.screen.getyx()
                self.screen.addstr(y, 0, "bandicoot> ")
                self.screen.addstr(y, len("bandicoot> "), self.history[-(history_index%len(self.history))])
                self.screen.clrtoeol()
                line = self.history[-(history_index%len(self.history))]
            elif s == curses.KEY_DOWN:
                if len(self.history) < 1:
                    # prevent divide by zero when history is 0
                    continue
                history_index -= 1
                cursor_offset = 0
                (y, x) = self.screen.getyx()
                self.screen.addstr(y, 0, "bandicoot> ")
                self.screen.addstr(y, len("bandicoot> "), self.history[-(history_index%len(self.history))])
                self.screen.clrtoeol()
                line = self.history[-(history_index%len(self.history))]
            elif s == curses.KEY_LEFT:
                if cursor_offset > len(line)*-1:
                    cursor_offset -= 1
                    (y, x) = self.screen.getyx()
                    self.screen.move(y, x-1)
            elif s == curses.KEY_RIGHT:
                if cursor_offset < 0:
                    cursor_offset += 1
                    (y, x) = self.screen.getyx()
                    self.screen.move(y, x+1)
            else:
                #self.screen.ddstr("Out of range: %d" % s)
                history_index = 0
                cursor_offset = 0

        curses.endwin()

    def blocking_get_response_queued_job(self, queue_id):
        global sig_bg_pressed
        global sig_kill_pressed
        sig_bg_pressed = 0 # Reset ctrl-z state
        sig_kill_pressed = 0 # reset ctrl-c state

        data = {"response": "  "}
        last_response = ""
        self.screen.addstr("\nJob is running with id=%s. Press ctrl-z to background job.\n" % str(queue_id))
        self.screen.refresh()

        while sig_bg_pressed == 0:
            if sig_kill_pressed == 1:
                # Kill job
                data = self.run_action(self.get_action_from_command("jobs kill id=%s" % str(queue_id)))
                self.screen.addstr(data["response"])
                self.screen.refresh()
                break
            else:
                data = self.run_action(self.get_action_from_command("jobs status id=%s" % str(queue_id)))
                if data is None or "finished" not in data or data["finished"] == True:
                    # Its finished, print the last update
                    updatestr = data["response"].replace(last_response, "")
                    self.screen.addstr(updatestr)
                    self.screen.refresh()
                    return "" # prints no update since everything was already printed to the screen
                if "response" in data and "exit_code" in data and data["exit_code"] != 0:
                    # Error happend
                    return data["response"] # prints error string
                updatestr = data["response"].replace(last_response, "")
                self.screen.addstr(updatestr)
                self.screen.refresh()
                last_response = data["response"]
                time.sleep(5)
        return "" # no update

    def shell_parse_line(self, line):
        line = line.strip()

        # Return nothing for an empty line
        if len(line) <= 0:
            return("")

        action = line.split()
        if self.is_action_quit(action):
            # bandicoot> quit
            # bandicoot> exit
            self.action_quit()
        else:
            # Server Side Handles Command Response
            # bandicoot> [category ..] action [option1=something ..]
            if line is not None and len(line) > 0:
                self.history.append(line)

            actionjson = self.get_action_from_command(line)
            if yacc.parser_error is None:
                data = self.run_action(actionjson)
            else:
                data = {"response": yacc.parser_error}
            if data is not None:
                if "response" in data:
                    return data["response"]
                elif "queue_id" in data:
                    return self.blocking_get_response_queued_job(data["queue_id"])
                else:
                    return("bandicoot - Invalid Response From server\n")
            else:
                return("bandicoot - Failed To Get Response From Server\n")

    def run(self):
        """ EntryPoint Of Application """
        self.login_prompt()
        if self.interactive_mode:
            curses.wrapper(self.startshell)
        else:
            for command in self.noninteractive_commands:
                print(self.shell_parse_line(command))




# parsetab.py
# This file is automatically generated. Do not edit.
_tabversion = '3.8'

_lr_method = 'LALR'

_lr_signature = '08D89100DBB39CDE1D5E28CD8ADD45C4'
    
_lr_action_items = {'OPTIONVALS':([8,],[13,]),'OPTIONVALD':([8,],[12,]),'SPACE':([1,3,5,6,7,10,11,12,13,14,],[-4,4,-6,-3,9,-7,-8,-10,-9,-5,]),'EQUAL':([6,15,],[8,8,]),'OPTIONVAL':([8,],[11,]),'ACTION':([0,4,8,9,],[1,6,10,15,]),'$end':([1,2,3,5,6,7,10,11,12,13,14,],[-4,0,-2,-6,-3,-1,-7,-8,-10,-9,-5,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'action_run':([0,],[2,]),'options':([4,],[7,]),'actions':([0,],[3,]),'option':([4,9,],[5,14,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> action_run","S'",1,None,None,None),
  ('action_run -> actions SPACE options','action_run',3,'p_action_run','yacc.py',16),
  ('action_run -> actions','action_run',1,'p_action_run','yacc.py',17),
  ('actions -> actions SPACE ACTION','actions',3,'p_actions','yacc.py',32),
  ('actions -> ACTION','actions',1,'p_actions','yacc.py',33),
  ('options -> options SPACE option','options',3,'p_options','yacc.py',48),
  ('options -> option','options',1,'p_options','yacc.py',49),
  ('option -> ACTION EQUAL ACTION','option',3,'p_option','yacc.py',61),
  ('option -> ACTION EQUAL OPTIONVAL','option',3,'p_option','yacc.py',62),
  ('option -> ACTION EQUAL OPTIONVALS','option',3,'p_option','yacc.py',63),
  ('option -> ACTION EQUAL OPTIONVALD','option',3,'p_option','yacc.py',64),
]



#!/usr/bin/env python

import os
import sys

sys.path.insert(0, os.path.abspath('lib'))
from bandicoot import __version__, __author__

try:
    from setuptools import setup, find_packages
except ImportError:
    print("bandicoot needs setuptools in order to build. Install it using"
            " your package manager (usually python-setuptools) or via pip (pip"
            " install setuptools).")
    sys.exit(1)

if len(sys.argv) > 1 and sys.argv[1] == 'bandicoot-cli':
    # python setup.py "bandicoot-cli" install
    del sys.argv[1]
    # bandicoot-cli only
    setup(name='bandicoot-cli',
          version=__version__,
          description='The command line and control center of a Data Center or Cloud.',
          author=__author__,
          author_email='david@davidwhiteside.com',
          url='https://github.com/starboarder2001/bandicoot',
          license='MIT',
          install_requires=["PyYAML", 'setuptools', 'ply', 'ascii_graph'],
          package_dir={ '': 'lib' },
          packages=find_packages('lib'),
          classifiers=[
              'Environment :: Console',
              'Intended Audience :: Developers',
              'Intended Audience :: Information Technology',
              'Intended Audience :: System Administrators',
              'License :: OSI Approved :: MIT License',
              'Natural Language :: English',
              'Operating System :: POSIX',
              'Programming Language :: Python :: 2.6',
              'Programming Language :: Python :: 2.7',
              'Topic :: System :: Installation/Setup',
              'Topic :: System :: Systems Administration',
              'Topic :: Utilities',
          ],
          scripts=[
             'bin/bandicoot',
             'bin/bandicoot-cli',
          ],
    )
else:
    # Full bandicoot Package
    # python setup.py install
    setup(name='bandicoots',
          version=__version__,
          description='The command line and control center of a Data Center or Cloud.',
          author=__author__,
          author_email='david@davidwhiteside.com',
          url='https://github.com/starboarder2001/bandicoot',
          license='MIT',
          install_requires=["requests", "Flask", "flask-cors", "PyJWT", "PyYAML", 'setuptools', 'ansible',
              'pymongo', 'ply', 'pycrypto', 'ldap3'],
          package_dir={ '': 'lib' },
          packages=find_packages('lib'),
          classifiers=[
              'Environment :: Console',
              'Intended Audience :: Developers',
              'Intended Audience :: Information Technology',
              'Intended Audience :: System Administrators',
              'License :: OSI Approved :: MIT License',
              'Natural Language :: English',
              'Operating System :: POSIX',
              'Programming Language :: Python :: 2.6',
              'Programming Language :: Python :: 2.7',
              'Topic :: System :: Installation/Setup',
              'Topic :: System :: Systems Administration',
              'Topic :: Utilities',
          ],
          scripts=[
             'bin/bandicoot',
             'bin/bandicoot-api',
             'bin/bandicoot-cli',
             'bin/bandicoot-api-install',
             'bin/bandicoot-api-install.yml',
          ],
          package_data={"": ["data_files/bandicoot-gui/*", "data_files/bandicoot-gui/css/*", "data_files/bandicoot-gui/imgs/*", "data_files/bandicoot-gui/js/*", "data_files/bandicoot-gui/templates/*"]},
    )
